#
# Copyright (c) 2019-2021 James Thorne.
#
# This file is part of factual error correction.
# See https://jamesthorne.co.uk for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import itertools
import json
import pickle
from logging import getLogger
from typing import Callable, Dict, Iterable, List

import numpy as np
from torch import nn
from torch.utils.data import Sampler
from transformers import BartTokenizer



def recursive_clean(metadata_dict):
    if isinstance(metadata_dict, dict):
        return {
            k: recursive_clean(v) for k, v in metadata_dict.items() if v is not None
        }
    elif isinstance(metadata_dict, list) or isinstance(metadata_dict, tuple):
        return [recursive_clean(k) for k in metadata_dict if k is not None]
    else:
        return metadata_dict


def encode_line(
    tokenizer, line, max_length, pad_to_max_length=True, return_tensors="pt"
):
    extra_kw = (
        {"add_prefix_space": True} if isinstance(tokenizer, BartTokenizer) else {}
    )
    return tokenizer(
        [line],
        max_length=max_length,
        padding="max_length" if pad_to_max_length else None,
        truncation=True,
        return_tensors=return_tensors,
        **extra_kw,
    )

def lmap(f: Callable, x: Iterable) -> List:
    """list(map(f, x))"""
    return list(map(f, x))


def trim_batch(
    input_ids,
    pad_token_id,
    attention_mask=None,
    token_type_ids=None,
):
    """Remove columns that are populated exclusively by pad_token_id"""
    keep_column_mask = input_ids.ne(pad_token_id).any(dim=0)
    if attention_mask is None and token_type_ids is None:
        return input_ids[:, keep_column_mask]
    elif attention_mask is not None and token_type_ids is None:
        return input_ids[:, keep_column_mask], attention_mask[:, keep_column_mask]
    else:
        return input_ids[:, keep_column_mask], attention_mask[:, keep_column_mask], token_type_ids[:, keep_column_mask]


class SortishSampler(Sampler):
    "Go through the text data by order of src length with a bit of randomness. From fastai repo."

    def __init__(self, data, batch_size):
        self.data, self.bs = data, batch_size

    def key(self, i):
        return self.data[i]

    def __len__(self) -> int:
        return len(self.data)

    def __iter__(self):
        idxs = np.random.permutation(len(self.data))
        sz = self.bs * 50
        ck_idx = [idxs[i : i + sz] for i in range(0, len(idxs), sz)]
        sort_idx = np.concatenate(
            [sorted(s, key=self.key, reverse=True) for s in ck_idx]
        )
        sz = self.bs
        ck_idx = [sort_idx[i : i + sz] for i in range(0, len(sort_idx), sz)]
        max_ck = np.argmax(
            [self.key(ck[0]) for ck in ck_idx]
        )  # find the chunk with the largest key,
        ck_idx[0], ck_idx[max_ck] = (
            ck_idx[max_ck],
            ck_idx[0],
        )  # then make sure it goes first.
        sort_idx = (
            np.concatenate(np.random.permutation(ck_idx[1:]))
            if len(ck_idx) > 1
            else np.array([], dtype=np.int)
        )
        sort_idx = np.concatenate((ck_idx[0], sort_idx))
        return iter(sort_idx)


logger = getLogger(__name__)


def use_task_specific_params(model, task):
    """Update config with summarization specific params."""
    task_specific_params = model.config.task_specific_params

    if task_specific_params is not None:
        pars = task_specific_params.get(task, {})
        logger.info(f"using task specific params for {task}: {pars}")
        model.config.update(pars)

def freeze_params(model: nn.Module):
    for par in model.parameters():
        par.requires_grad = False


def grad_status(model: nn.Module) -> Iterable:
    return (par.requires_grad for par in model.parameters())


def any_requires_grad(model: nn.Module) -> bool:
    return any(grad_status(model))


def assert_all_frozen(model):
    model_grads: List[bool] = list(grad_status(model))
    n_require_grad = sum(lmap(int, model_grads))
    npars = len(model_grads)
    assert not any(
        model_grads
    ), f"{n_require_grad/npars:.1%} of {npars} weights require grad"


def assert_not_all_frozen(model):
    model_grads: List[bool] = list(grad_status(model))
    npars = len(model_grads)
    assert any(model_grads), f"none of {npars} weights require grad"


def is_truthy(arg):
    return arg.strip().lower() in ["1", "y", "yes", "t", "true"]





import numpy as np
from abc import ABC

from transformers import EvalPrediction

from xfact.registry.registrable import Registrable


class Scorer(Registrable, ABC):
    def __call__(self, actual, predicted, **kwargs):
        raise NotImplementedError()


@Scorer.register("multiset_information_retrieval")
class InformationRetrievalScorer(Scorer):

    def __call__(self, actual, predicted, **kwargs):
        return {
            "macro_recall": macro(max_over_many(recall), actual, predicted),
            "macro_precision": macro(max_over_many(precision), actual, predicted),
            "macro_f1": macro(max_over_many(f1), actual, predicted),
            "macro_r_precision": macro(max_over_many(r_precision), actual, predicted),
            "macro_average_precision": macro(max_over_many(average_precision), actual, predicted),
            "macro_recall_corrected": macro(max_over_many(recall_corrected), actual, predicted),
            "macro_precision_corrected": macro(max_over_many(precision_corrected), actual, predicted),
            "macro_reciprocal_rank": macro(max_over_many(reciprocal_rank), actual, predicted),
            "macro_average_precision_corrected": macro(max_over_many(average_precision_corrected), actual, predicted)
        }


@Scorer.register("information_retrieval")
class InformationRetrievalScorer(Scorer):

    def __call__(self, actual, predicted, **kwargs):
        rec = macro(recall, actual, predicted)
        pr = macro(precision, actual, predicted)
        return {
            "macro_recall": rec,
            "macro_precision": pr,
            "macro_f1": f1(pr,rec),
            "macro_r_precision": macro(r_precision, actual, predicted),
        }


@Scorer.register("classification")
class ClassificationScorer(Scorer):
    def __call__(self, actual, predicted, **kwargs):
        return {
            # "macro_recall": macro(recall, actual, predicted),
            # "macro_precision": macro(precision, actual, predicted),
            # "macro_f1": macro(f1, actual, predicted),
            "em": macro(exact_match, actual, predicted)
        }






@Scorer.register("qa")
class ClassificationScorer(Scorer):
    def __call__(self, actual, predicted, **kwargs):
        return {
            "em": macro(exact_match, actual, predicted)
        }


def exact_match(actual, predicted):
    return 1.0 if actual == predicted else 0.0


def lrp(actual, predicted):
    actual = list(actual)
    predicted = list(predicted)
    rank = 1
    denom = 1
    found = False
    if len(predicted) and len(actual):
        for p in reversed(predicted):
            if p in actual and not found:
                found = True
            elif p not in actual and found:
                rank += 1
                denom += 1
            elif p not in actual:
                denom += 1
    return found, rank, denom


def precision(actual, predicted):
    actual = set(actual)
    predicted = set(predicted)
    return (
        sum(1.0 for p in predicted if p in actual) / float(len(predicted))
        if len(predicted)
        else 1.0
    )


def precision_corrected(actual, predicted):
    found, _, denom = lrp(actual, predicted)
    return 0 if not found else 1. / denom


def precision_at_k(r, k):
    assert k >= 1
    r = np.asarray(r)[:k] != 0
    if r.size != k:
        raise ValueError('Relevance score length < k')
    return np.mean(r)


def average_precision(actual, predicted):
    """
    https://gist.github.com/bwhite/3726239
    """
    r = []
    for i, p in enumerate(predicted):
        if len(actual) > i and actual[i] == predicted[i]:
            r.append(1)
        else:
            r.append(0)
    r = np.asarray(r) != 0
    out = [precision_at_k(r, k + 1) for k in range(r.size) if r[k]]
    if not out:
        return 0.
    return np.mean(out)


def average_precision_corrected(actual, predicted):
    """
    https://gist.github.com/bwhite/3726239
    """
    r = []
    found = False
    predicted_len = len(predicted) - 1
    for j, _ in enumerate(predicted):
        i = predicted_len - j
        if len(actual) > i and actual[i] == predicted[i] and not found:
            found = True
            r.insert(0, 1)
        elif len(actual) > i and actual[i] == predicted[i]:
            continue
        else:
            r.insert(0, 0)
    r = np.asarray(r) != 0
    out = [precision_at_k(r, k + 1) for k in range(r.size) if r[k]]
    if not out:
        return 0.
    return np.mean(out)


def reciprocal_rank(actual, predicted):
    """
    https://machinelearning.wtf/terms/mean-reciprocal-rank-mrr/
    I assumed that the lists of evidences in the 1 index of evidences matrix were not sorted by relevance score.
    This program is called with one query, so this program takes maximum of
    reciprocal rank of elements in the actual
    """
    found, rank, _ = lrp(actual, predicted)
    return 0 if not found else 1. / rank


def recall(actual, predicted):
    actual = set(actual)
    predicted = set(predicted)
    return (
        sum(1.0 for a in actual if a in predicted) / float(len(actual))
        if len(actual)
        else 1.0
    )


def recall_corrected(actual, predicted):
    found, _, denom = lrp(predicted, actual)
    return 0 if not found else 1. / denom


def r_precision(actual, predicted):
    actual = set(actual)

    R = len(actual)
    r_predicted = predicted[:R]

    return (
        sum(1.0 for p in r_predicted if p in actual) / float(R)
        if R
        else 1.0
    )


def f1(prec, rec):
    return 2*prec*rec/(prec+rec) if prec+rec > 0 else 0


def macro(method, actual, predicted):
    return sum(method(a,p) for a, p in zip(actual,predicted) if a is not None) / len([a for a in actual if a is not None])


def max_over_many(method):
    def apply_multi(actual_many, predicted):
        scores = []
        for actual in actual_many:
            scores.append(method(actual, predicted))
        return max(scores)
    return apply_multi





import importlib
import pkgutil
import logging

logger = logging.getLogger(__name__)



def import_submodules(package_name: str) -> None:
    importlib.invalidate_caches()
    logger.info(f"Import packages from {package_name}")

    module = importlib.import_module(package_name)
    path = getattr(module, '__path__', [])
    path_string = '' if not path else path[0]

    # walk_packages only finds immediate children, so need to recurse.
    for module_finder, name, _ in pkgutil.walk_packages(path):
        # Sometimes when you import third-party libraries that are on your path,
        # `pkgutil.walk_packages` returns those too, so we need to skip them.
        if path_string and module_finder.path != path_string:
            continue
        subpackage = f"{package_name}.{name}"
        import_submodules(subpackage)
        logger.info(f"Import subpackage {subpackage}")




import logging
import operator
from abc import ABC
from collections import defaultdict, Counter

import torch
from torch.utils.data import Dataset as TorchDataset
from tqdm import tqdm
from typing import Dict
from pathlib import Path
from xfact.nlp.data_utils import encode_line, trim_batch, SortishSampler
from xfact.registry.registrable import Registrable
from overrides import overrides
logger = logging.getLogger(__name__)


class XFactDataset(TorchDataset, Registrable, ABC):
    def __init__(
            self,
            tokenizer,
            instance_generator,
            max_source_length,
            name="",
            n_obs=None,
            test_mode=False,
            output_prompt="",
            streaming=False,
            num_instances_to_preview=5,
            sep_token="<sep />"
    ):
        super().__init__()


        # Default sep token. If tokenizer has a sep token, this will be ignored.
        self.sep_token = tokenizer._sep_token or sep_token

        # Set class specific things
        self.max_source_length = max_source_length
        self.tokenizer = tokenizer

        # We want to preview a number of instances
        self.num_instances_to_preview = num_instances_to_preview
        self.blind_test_mode = test_mode
        self.has_preview = num_instances_to_preview

        # Set pad token
        self.pad_token_id = self.tokenizer.pad_token_id

        # Set sep token
        if not self.sep_token in self.tokenizer.vocab:
            logger.info("Tokenizer doesn't have a sep token. Create it")
            self.tokenizer.add_tokens([self.sep_token], special_tokens=True)
            self.tokenizer._sep_token = self.sep_token

        self.sep_token_id = self.tokenizer.vocab.get(self.sep_token)
        self.sep_token = self.tokenizer._sep_token
        logger.info(f"Sep token id is {self.sep_token_id}")

        # If we're streaming instances from disk, then we should continuously generate instances
        # Otherwise, load instances from an instance generator
        self.streaming = streaming
        if not self.streaming:
            self.instances = list(tqdm(filter(lambda i: i is not None, instance_generator), desc=name))
            self.generated = list(tqdm(map(self.generate, self.instances), desc="Generating instances"))
        else:
            self.instances = filter(lambda i: i is not None, instance_generator)

        # Needed for Sortish sampler
        if n_obs is not None:
            self.src_lens = self.src_lens[:n_obs]

    def __len__(self):
        return len(self.generated)

    def prepare_src(self, instance):
        raise NotImplementedError()

    def prepare_tgt(self, instance):
        raise NotImplementedError()

    def generate(self, instance) -> Dict[str, torch.Tensor]:
        raise NotImplementedError()

    def __getitem__(self, index) -> Dict[str, torch.Tensor]:
        instance = self.generated[index] if not self.streaming else \
            self.generate(next(self.instances))

        return instance

    @staticmethod
    def get_char_lens(data_file):
        return [len(x) for x in Path(data_file).open().readlines()]

    @staticmethod
    def trim_seq2seq_batch(batch, pad_token_id) -> tuple:
        y = trim_batch(batch["decoder_input_ids"], pad_token_id)
        source_ids, source_mask = trim_batch(
            batch["input_ids"], pad_token_id, attention_mask=batch["attention_mask"]
        )
        return source_ids, source_mask, y

    @staticmethod
    def collate_fn(model, batch, pad_token_id, ignore_pad_token_for_loss=True) -> Dict[str, torch.Tensor]:
        input_ids = torch.stack([x["input_ids"] for x in batch])
        masks = torch.stack([x["attention_mask"] for x in batch])

        source_ids, source_mask = trim_batch(
            input_ids, pad_token_id, attention_mask=masks
        )

        ret_batch = {
            "input_ids": source_ids,
            "attention_mask": source_mask,
        }

        if "decoder_input_ids" in batch[0]:
            target_ids = torch.stack([x["decoder_input_ids"] for x in batch])
            y = trim_batch(target_ids, -100 if ignore_pad_token_for_loss else pad_token_id)
            ret_batch["labels"] = y

            # prepare decoder_input_ids
            if (
                    ret_batch['labels'] is not None
                    and model is not None
                    and hasattr(model, "prepare_decoder_input_ids_from_labels")
            ):
                decoder_input_ids = model.prepare_decoder_input_ids_from_labels(labels=ret_batch["labels"])
                ret_batch["decoder_input_ids"] = decoder_input_ids

            # ret_batch["decoder_input_ids"] = y

        return ret_batch

    def make_sortish_sampler(self, batch_size):
        return SortishSampler(self.src_lens, batch_size)


class XFactSeq2SeqDataset(XFactDataset, ABC):
    def __init__(
            self,
            tokenizer,
            instance_generator,
            max_source_length,
            max_target_length=32,
            name="",
            n_obs=None,
            test_mode=False,
            output_prompt="",
            streaming=False
    ):
        self.max_target_length = max_target_length
        self.output_prompt = output_prompt
        self.prompt_tokens = tokenizer(self.output_prompt)['input_ids'][:-1]

        super(XFactSeq2SeqDataset, self).__init__(tokenizer=tokenizer,
                                                  instance_generator=instance_generator,
                                                  max_source_length=max_source_length,
                                                  name=name,
                                                  n_obs=n_obs,
                                                  test_mode=test_mode,
                                                  output_prompt=output_prompt,
                                                  streaming=streaming)

        logger.info(f"Output prompt tokens are {self.prompt_tokens}")

    @staticmethod
    def collate_fn(model, batch, pad_token_id, ignore_pad_token_for_loss=True) -> Dict[str, torch.Tensor]:
        input_ids = torch.stack([x["input_ids"] for x in batch])
        masks = torch.stack([x["attention_mask"] for x in batch])

        source_ids, source_mask = trim_batch(
            input_ids, pad_token_id, attention_mask=masks
        )

        ret_batch = {
            "input_ids": source_ids,
            "attention_mask": source_mask,
        }

        if "decoder_input_ids" in batch[0]:
            target_ids = torch.stack([x["decoder_input_ids"] for x in batch])
            y = trim_batch(target_ids, -100 if ignore_pad_token_for_loss else pad_token_id)
            ret_batch["labels"] = y

            # prepare decoder_input_ids
            if (
                    ret_batch['labels'] is not None
                    and model is not None
                    and hasattr(model, "prepare_decoder_input_ids_from_labels")
            ):
                decoder_input_ids = model.prepare_decoder_input_ids_from_labels(labels=ret_batch["labels"])
                ret_batch["decoder_input_ids"] = decoder_input_ids

            # ret_batch["decoder_input_ids"] = y

        return ret_batch

    def generate(self, instance) -> Dict[str, torch.Tensor]:
        source_input = self.prepare_src(instance)
        source_inputs = encode_line(
            self.tokenizer, source_input, self.max_source_length
        )

        source_ids = source_inputs["input_ids"].squeeze()
        src_mask = source_inputs["attention_mask"].squeeze()

        if not self.blind_test_mode:
            target_input = self.prepare_tgt(instance)
            target_inputs = encode_line(
                self.tokenizer, target_input, self.max_target_length
            )
            target_ids = target_inputs["input_ids"].squeeze()

        if self.has_preview >= 0:
            self.has_preview -= 1
            print(source_input)
            print(source_ids)

            if not self.blind_test_mode:
                print(target_input)
                print(target_ids)

            print("*" * 100)

        ret = {
            "input_ids": source_ids,
            "attention_mask": src_mask
        }

        if not self.blind_test_mode:
            ret["decoder_input_ids"] = target_ids

        return ret


class XFactTaggingDataset(XFactDataset):
    pass


class XFactClassificationDataset(XFactDataset, ABC):
    def __init__(self, tokenizer, instance_generator, max_source_length, label_dict=None, **kwargs):
        self.label_dict = defaultdict(int) if label_dict is None else label_dict
        super().__init__(tokenizer, instance_generator, max_source_length,**kwargs)
        self.class_weights = self.get_label_distribution()

    def get_label_distribution(self):
        label_counts = Counter(map(operator.itemgetter("label"), self.instances))

        logger.info(f"Class sizes: {label_counts}")
        summed = sum(label_counts.values())
        return {k:v/summed for k,v in label_counts.items()}


    def prepare_tgt(self, instance):
        return instance["label"]

    @overrides
    def generate(self, instance) -> Dict[str, torch.Tensor]:
        # instance = self.instances[index] if not self.streaming else next(self.instances)

        source_input = self.prepare_src(instance)
        source_inputs = encode_line(
            self.tokenizer, source_input, self.max_source_length
        )

        source_ids = source_inputs["input_ids"].squeeze()
        src_mask = source_inputs["attention_mask"].squeeze()
        src_types = source_inputs["token_type_ids"].squeeze() if "token_type_ids" in source_inputs else None

        if not self.blind_test_mode:
            target_input = self.prepare_tgt(instance)

            if target_input not in self.label_dict:
                self.label_dict[target_input] = len(self.label_dict)
            target_ids = torch.LongTensor([self.label_dict[target_input]])


        if self.num_instances_to_preview >= 0:
            self.num_instances_to_preview -= 1
            print(source_input)
            print(source_ids)

            if not self.blind_test_mode:
                print(target_input)
                print(target_ids)

            print("*" * 100)

        ret = {
            "input_ids": source_ids,
            "attention_mask": src_mask,
            "token_type_ids": src_types
        }

        if not self.blind_test_mode:
            ret["label_ids"] = target_ids

        return ret

    @staticmethod
    def collate_fn(model, batch, pad_token_id, ignore_pad_token_for_loss=True) -> Dict[str, torch.Tensor]:
        input_ids = torch.stack([x["input_ids"] for x in batch])
        masks = torch.stack([x["attention_mask"] for x in batch])

        # types = torch.stack([x["token_type_ids"] for x in batch]) if "token_type_ids" in batch[0] and batch[0]["token_type_ids"] is not None else None

        source_ids, source_mask = trim_batch(
            input_ids, pad_token_id, attention_mask=masks
        )

        # source_ids, source_mask = input_ids,masks
        #     # ,types
        ret_batch = {
            "input_ids": source_ids,
            "attention_mask": source_mask,
            # "token_type_ids": source_types
        }

        if "label_ids" in batch[0]:
            target_ids = torch.stack([x["label_ids"] for x in batch])
            ret_batch["labels"] = target_ids


            # ret_batch["decoder_input_ids"] = y

        return ret_batch
